//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2018 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include "stdafx.h"
#include <at/atcore/devicesio.h>
#include <at/atcore/wraptime.h>
#include <at/atcpu/history.h>
#include <at/atcpu/memorymap.h>
#include "diskdrivefullbase.h"
#include "firmwaremanager.h"

ATDiskDriveAudioPlayer::ATDiskDriveAudioPlayer() {
}

ATDiskDriveAudioPlayer::~ATDiskDriveAudioPlayer() {
	VDASSERT(!mpAudioMixer);
}

void ATDiskDriveAudioPlayer::Shutdown() {
	if (mpAudioMixer) {
		SetRotationSoundEnabled(false);
		mpAudioMixer = nullptr;
	}
}

void ATDiskDriveAudioPlayer::SetRotationSoundEnabled(bool enabled) {
	if (!mpAudioMixer)
		return;

	if (enabled) {
		if (mRotationSoundId == ATSoundId::Invalid) {
			mRotationSoundId = mpAudioMixer->GetSamplePlayer().AddLoopingSound(kATAudioMix_Drive, 0, kATAudioSampleId_DiskRotation, 1.0f);
		}
	} else {
		if (mRotationSoundId != ATSoundId::Invalid) {
			mpAudioMixer->GetSamplePlayer().StopSound(mRotationSoundId);
			mRotationSoundId = ATSoundId::Invalid;
		}
	}
}

void ATDiskDriveAudioPlayer::PlayStepSound(ATAudioSampleId sampleId, float volume) {
	if (!mpAudioMixer)
		return;

	mpAudioMixer->GetSamplePlayer().AddSound(kATAudioMix_Drive, 0, sampleId, volume);
}

void ATDiskDriveAudioPlayer::InitAudioOutput(IATAudioMixer *mixer) {
	mpAudioMixer = mixer;
}

///////////////////////////////////////////////////////////////////////////

ATDiskDriveChangeHandler::ATDiskDriveChangeHandler()
	: mpUpdateWriteProtect([](auto) {})
	, mpUpdateDiskReady([](auto) {})
{
}

ATDiskDriveChangeHandler::~ATDiskDriveChangeHandler() {
	Shutdown();
}

void ATDiskDriveChangeHandler::Init(ATScheduler *scheduler) {
	static constexpr float kDiskChangeStepMS = 50.0f;

	mpScheduler = scheduler;
	mCyclesPerStep = (uint32_t)(0.5f + scheduler->GetRate().asDouble() * (kDiskChangeStepMS / 1000.0f));
}

void ATDiskDriveChangeHandler::Shutdown() {
	if (mpScheduler) {
		mpScheduler->UnsetEvent(mpEvent);
		mpScheduler = nullptr;
	}
}

void ATDiskDriveChangeHandler::SetOutputStateFns(vdfunction<void(std::optional<bool>)> wpOverrideChange, vdfunction<void(std::optional<bool>)> readyOverrideChange) {
	mpUpdateWriteProtect = std::move(wpOverrideChange);
	mpUpdateDiskReady = std::move(readyOverrideChange);
}

void ATDiskDriveChangeHandler::Reset() {
	if (mpScheduler)
		mpScheduler->UnsetEvent(mpEvent);

	mDiskChangeState = 0;
	ForceOutputUpdate();
}

void ATDiskDriveChangeHandler::ChangeDisk() {
	mDiskChangeState = 0;
	mpScheduler->SetEvent(1, this, 1, mpEvent);
}

void ATDiskDriveChangeHandler::ForceOutputUpdate() const {
	mpUpdateWriteProtect(GetWriteProtectState());
	mpUpdateDiskReady(GetDiskReadyState());
}

void ATDiskDriveChangeHandler::OnScheduledEvent(uint32 id) {
	mpEvent = nullptr;

	if (++mDiskChangeState >= 4)
		mDiskChangeState = 0;
	else
		mpScheduler->SetEvent(mCyclesPerStep, this, 1, mpEvent);

	mpUpdateWriteProtect(GetWriteProtectState());

	switch(mDiskChangeState) {
		case 1:
		case 0:
			mpUpdateDiskReady(GetDiskReadyState());
			break;

		case 2:
		case 3:
			break;
	}
}

///////////////////////////////////////////////////////////////////////////

ATDiskDriveSerialQueue::ATDiskDriveSerialQueue()
	: mpOnDriveCommandStateChanged([](bool) {})
{
}

ATDiskDriveSerialQueue::~ATDiskDriveSerialQueue() {
	Shutdown();
}

void ATDiskDriveSerialQueue::Init(ATScheduler *masterScheduler, ATScheduler *driveScheduler, IATDeviceSIOManager *sioMgr) {
	mpSIOMgr = sioMgr;
	mpMasterScheduler = masterScheduler;
	mpDriveScheduler = driveScheduler;
}

void ATDiskDriveSerialQueue::Shutdown() {
	Reset();
	mpMasterScheduler = nullptr;
	mpDriveScheduler = nullptr;
	mpSIOMgr = nullptr;
}

void ATDiskDriveSerialQueue::SetOnDriveCommandStateChanged(vdfunction<void(bool)> fn) {
	mpOnDriveCommandStateChanged = std::move(fn);
}

void ATDiskDriveSerialQueue::Reset() {
	mCommandQueue.clear();
	mTransmitHead = 0;
	mTransmitTail = 0;

	mbDriveCommandState = false;

	if (mpMasterScheduler) {
		mpMasterScheduler->UnsetEvent(mpEventTransmit);
	}

	if (mpDriveScheduler) {
		mpDriveScheduler->UnsetEvent(mpEventDriveCommandChange);
	}
}

void ATDiskDriveSerialQueue::AddTransmitByte(uint32 masterTime, uint8 data, uint32 cyclesPerBit) {
	VDASSERT(cyclesPerBit);
	static_assert(!(kTransmitQueueSize & (kTransmitQueueSize - 1)), "mTransmitQueue size not pow2");

	// check if previous transition is at same time
	const uint32 queueLen = (mTransmitTail - mTransmitHead) & kTransmitQueueMask;
	if (queueLen) {
		auto& prevEdge = mTransmitQueue[(mTransmitTail - 1) & kTransmitQueueMask];

		// check if this event is at or before the last event in the queue
		if (!((prevEdge.mTime - masterTime) & UINT32_C(0x80000000))) {
			// check if we've gone backwards in time and drop event if so
			if (prevEdge.mTime == masterTime) {
				// same time -- overwrite event
				prevEdge.mData = data;
				prevEdge.mCyclesPerBit = cyclesPerBit;
			} else {
				VDASSERT(!"Dropping new event earlier than tail in transmit queue.");
			}

			return;
		}
	}

	// check if we have room for a new event
	if (queueLen >= kTransmitQueueMask) {
		VDASSERT(!"Transmit queue full.");
		return;
	}

	// add the new event
	auto& newEdge = mTransmitQueue[mTransmitTail++ & kTransmitQueueMask];

	newEdge.mTime = masterTime;
	newEdge.mData = data;
	newEdge.mCyclesPerBit = cyclesPerBit;

	// queue next event if needed
	QueueNextTransmitEvent();
}

void ATDiskDriveSerialQueue::OnScheduledEvent(uint32 id) {
	if (id == kEventId_Transmit) {
		mpEventTransmit = nullptr;

		OnTransmitEvent();
	} else if (id == kEventId_DriveCommandChange) {
		mpEventDriveCommandChange = nullptr;
		OnCommandChangeEvent();
	}
}

void ATDiskDriveSerialQueue::OnTransmitEvent() {
	// drain transmit queue entries until we're up to date
	const uint32 t = ATSCHEDULER_GETTIME(mpMasterScheduler);

	while((mTransmitHead ^ mTransmitTail) & kTransmitQueueMask) {
		const auto& nextEdge = mTransmitQueue[mTransmitHead & kTransmitQueueMask];

		if ((t - nextEdge.mTime) & UINT32_C(0x80000000))
			break;

		mpSIOMgr->SendRawByte(nextEdge.mData, nextEdge.mCyclesPerBit, false, false, true);

		++mTransmitHead;
	}

	QueueNextTransmitEvent();
}

void ATDiskDriveSerialQueue::QueueNextTransmitEvent() {
	// exit if we already have an event queued
	if (mpEventTransmit)
		return;

	// exit if transmit queue is empty
	if (!((mTransmitHead ^ mTransmitTail) & kTransmitQueueMask))
		return;

	const auto& nextEvent = mTransmitQueue[mTransmitHead & kTransmitQueueMask];
	const uint32 delta = nextEvent.mTime - ATSCHEDULER_GETTIME(mpMasterScheduler);

	mpEventTransmit = mpMasterScheduler->AddEvent((uint32)(delta - 1) & UINT32_C(0x80000000) ? 1 : delta, this, kEventId_Transmit);
}

void ATDiskDriveSerialQueue::AddCommandEdge(uint32 driveTime, uint32 polarity) {
	// check if previous transition is at same time
	while(!mCommandQueue.empty()) {
		auto& prevEdge = mCommandQueue.back();

		// check if this event is at or before the last event in the queue
		if ((prevEdge.mTime - driveTime) & UINT32_C(0x80000000)) {
			// The previous edge is before the time of this edge, so we're good.
			break;
		}

		// check if we've gone backwards in time and drop event if so
		if (prevEdge.mTime == driveTime) {
			// same time -- overwrite event and exit
			prevEdge.mBit = polarity;
			return;
		}
			
		if (polarity && !prevEdge.mBit) {
			// If we're asserting /COMMAND, allow it to supercede earlier deassert events.
			mCommandQueue.pop_back();
		} else {
			VDASSERT(!"Dropping new event earlier than tail in command change queue.");
			return;
		}
	}

	// add the new event
	mCommandQueue.push_back( { driveTime, polarity } );

	// queue next event if needed
	QueueNextCommandEvent();
}

void ATDiskDriveSerialQueue::QueueNextCommandEvent() {
	// exit if we already have an event queued
	if (mpEventDriveCommandChange)
		return;

	// exit if transmit queue is empty
	if (mCommandQueue.empty())
		return;

	const auto& nextEvent = mCommandQueue.front();
	uint32 delta = (nextEvent.mTime - ATSCHEDULER_GETTIME(mpDriveScheduler)) & UINT32_C(0x7FFFFFFF);

	mpEventDriveCommandChange = mpDriveScheduler->AddEvent((uint32)(delta - 1) & UINT32_C(0x40000000) ? 1 : delta, this, kEventId_DriveCommandChange);
}

void ATDiskDriveSerialQueue::OnCommandChangeEvent() {
	// drain command queue entries until we're up to date
	const uint32 t = ATSCHEDULER_GETTIME(mpDriveScheduler);

	while(!mCommandQueue.empty()) {
		const auto& nextEdge = mCommandQueue.front();

		if ((t - nextEdge.mTime) & UINT32_C(0x40000000))
			break;

		bool bit = (nextEdge.mBit != 0);

		if (mbDriveCommandState != bit) {
			mbDriveCommandState = bit;

			mpOnDriveCommandStateChanged(bit);
		}

		mCommandQueue.pop_front();
	}

	// if we still have events, queue another event
	if (!mCommandQueue.empty()) {
		mpMasterScheduler->SetEvent(mCommandQueue.front().mTime - t, this, kEventId_DriveCommandChange, mpEventDriveCommandChange);
	}
}

///////////////////////////////////////////////////////////////////////////

void ATDiskDriveFirmwareControl::Init(void *mem, uint32 len, ATFirmwareType type) {
	mpFirmware = mem;
	mFirmwareLen = len;
	mFirmwareType = type;
}

void ATDiskDriveFirmwareControl::Shutdown() {
	mpFwMgr = nullptr;
}

void ATDiskDriveFirmwareControl::InitFirmware(ATFirmwareManager *fwman) {
	mpFwMgr = fwman;

	ReloadFirmware();
}

bool ATDiskDriveFirmwareControl::ReloadFirmware() {
	const uint64 id = mpFwMgr->GetFirmwareOfType(mFirmwareType, true);
	
	uint32 len = 0;
	bool changed = false;
	mpFwMgr->LoadFirmware(id, mpFirmware, 0, mFirmwareLen, &changed, &len, nullptr, nullptr, &mbFirmwareUsable);

	return changed;
}

const wchar_t *ATDiskDriveFirmwareControl::GetWritableFirmwareDesc(uint32 idx) const {
	return nullptr;
}

bool ATDiskDriveFirmwareControl::IsWritableFirmwareDirty(uint32 idx) const {
	return false;
}

void ATDiskDriveFirmwareControl::SaveWritableFirmware(uint32 idx, IVDStream& stream) {
}

ATDeviceFirmwareStatus ATDiskDriveFirmwareControl::GetFirmwareStatus() const {
	return mbFirmwareUsable ? ATDeviceFirmwareStatus::OK : ATDeviceFirmwareStatus::Missing;
}

///////////////////////////////////////////////////////////////////////////

void ATDiskDriveDebugTargetControl::InitTargetControl(IATDiskDriveDebugTargetProxy& proxy, double timestampFrequency, ATDebugDisasmMode disasmMode, ATDebugTargetBreakpointsBase *bpImpl) {
	mpProxy = &proxy;
	mTimestampFrequency = timestampFrequency;
	mpBreakpointsImpl = bpImpl;
	mDisasmMode = disasmMode;

	const auto rwmaps = proxy.GetReadWriteMaps();
	mpReadMap = rwmaps.first;
	mpWriteMap = rwmaps.second;
}

void *ATDiskDriveDebugTargetControl::AsInterface(uint32 iid) {
	switch(iid) {
		case IATDeviceDebugTarget::kTypeID: return static_cast<IATDeviceDebugTarget *>(this);
		case IATDebugTargetBreakpoints::kTypeID: return static_cast<IATDebugTargetBreakpoints *>(mpBreakpointsImpl);
		case IATDebugTargetHistory::kTypeID: return static_cast<IATDebugTargetHistory *>(this);
		case IATDebugTargetExecutionControl::kTypeID: return static_cast<IATDebugTargetExecutionControl *>(this);

		default:
			return nullptr;
	}
}

IATDebugTarget *ATDiskDriveDebugTargetControl::GetDebugTarget(uint32 index) {
	if (index == 0)
		return this;

	return nullptr;
}

const char *ATDiskDriveDebugTargetControl::GetName() {
	return "Disk Drive CPU";
}

ATDebugDisasmMode ATDiskDriveDebugTargetControl::GetDisasmMode() {
	return mDisasmMode;
}

void ATDiskDriveDebugTargetControl::GetExecState(ATCPUExecState& state) {
	mpProxy->GetExecState(state);
}

void ATDiskDriveDebugTargetControl::SetExecState(const ATCPUExecState& state) {
	mpProxy->SetExecState(state);
}

uint8 ATDiskDriveDebugTargetControl::ReadByte(uint32 address) {
	return DebugReadByte(address);
}

void ATDiskDriveDebugTargetControl::ReadMemory(uint32 address, void *dst, uint32 n) {
	return DebugReadMemory(address, dst, n);
}

uint8 ATDiskDriveDebugTargetControl::DebugReadByte(uint32 address) {
	uint8 v;
	ATCoProcReadMemory(mpReadMap, &v, address, 1);

	return v;
}

void ATDiskDriveDebugTargetControl::DebugReadMemory(uint32 address, void *dst, uint32 n) {
	ATCoProcReadMemory(mpReadMap, dst, address, n);
}

void ATDiskDriveDebugTargetControl::WriteByte(uint32 address, uint8 value) {
	ATCoProcWriteMemory(mpWriteMap, &value, address, 1);
}

void ATDiskDriveDebugTargetControl::WriteMemory(uint32 address, const void *src, uint32 n) {
	ATCoProcWriteMemory(mpWriteMap, src, address, n);
}

bool ATDiskDriveDebugTargetControl::GetHistoryEnabled() const {
	return !mHistory.empty();
}

void ATDiskDriveDebugTargetControl::SetHistoryEnabled(bool enable) {
	if (enable) {
		if (mHistory.empty()) {
			mHistorySize = 0x20000;
			mHistoryMask = mHistorySize - 1;

			mHistory.resize(mHistorySize, ATCPUHistoryEntry());
			mpProxy->SetHistoryBuffer(mHistory.data());
		}
	} else {
		if (!mHistory.empty()) {
			decltype(mHistory) tmp;
			tmp.swap(mHistory);
			mHistory.clear();
			mpProxy->SetHistoryBuffer(nullptr);
		}
	}
}

std::pair<uint32, uint32> ATDiskDriveDebugTargetControl::GetHistoryRange() const {
	const uint32 hcnt = mpProxy->GetHistoryCounter();

	return std::pair<uint32, uint32>(hcnt - mHistorySize, hcnt);
}

uint32 ATDiskDriveDebugTargetControl::ExtractHistory(const ATCPUHistoryEntry **hparray, uint32 start, uint32 n) const {
	if (!n || mHistory.empty())
		return 0;

	const ATCPUHistoryEntry *hstart = mHistory.data();
	const ATCPUHistoryEntry *hend = hstart + mHistorySize;
	const ATCPUHistoryEntry *hsrc = hstart + (start & mHistoryMask);

	for(uint32 i=0; i<n; ++i) {
		*hparray++ = hsrc;

		if (++hsrc == hend)
			hsrc = hstart;
	}

	return n;
}

double ATDiskDriveDebugTargetControl::GetTimestampFrequency() const {
	return mTimestampFrequency;
}

void ATDiskDriveDebugTargetControl::Break() {
	CancelStep();
}

bool ATDiskDriveDebugTargetControl::StepInto(const vdfunction<void(bool)>& fn) {
	CancelStep();

	mpStepHandler = fn;
	mbStepOut = false;
	mStepStartTime = mpProxy->GetTime();
	mpBreakpointsImpl->SetStepActive(true);
	Sync();
	return true;
}

bool ATDiskDriveDebugTargetControl::StepOver(const vdfunction<void(bool)>& fn) {
	CancelStep();

	mpStepHandler = fn;
	mbStepOut = true;
	mStepStartTime = mpProxy->GetTime();
	mStepOutSP = mpProxy->GetStepStackLevel();
	mpBreakpointsImpl->SetStepActive(true);
	Sync();
	return true;
}

bool ATDiskDriveDebugTargetControl::StepOut(const vdfunction<void(bool)>& fn) {
	CancelStep();

	mpStepHandler = fn;
	mbStepOut = true;
	mStepStartTime = mpProxy->GetTime();
	mStepOutSP = mpProxy->GetStepStackLevel() + 1;
	mpBreakpointsImpl->SetStepActive(true);
	Sync();
	return true;
}

void ATDiskDriveDebugTargetControl::StepUpdate() {
	Sync();
}

void ATDiskDriveDebugTargetControl::RunUntilSynced() {
	CancelStep();
	Sync();
}

bool ATDiskDriveDebugTargetControl::CheckBreakpoint(uint32 pc) {
	if (mpProxy->GetTime() == mStepStartTime)
		return false;

	const bool bpHit = mpBreakpointsImpl->CheckBP(pc);

	if (!bpHit) {
		if (mbStepOut) {
			// Keep stepping if wrapped(s < s0).
			if ((mpProxy->GetStepStackLevel() - mStepOutSP) & 0x80000000)
				return false;
		}
	}

	mpBreakpointsImpl->SetStepActive(false);

	mbStepNotifyPending = true;
	mbStepNotifyPendingBP = bpHit;
	return true;
}

void ATDiskDriveDebugTargetControl::CancelStep() {
	if (mpStepHandler) {
		mpBreakpointsImpl->SetStepActive(false);

		auto p = std::move(mpStepHandler);
		mpStepHandler = nullptr;

		p(false);
	}
}

void ATDiskDriveDebugTargetControl::FlushStepNotifications2() {
	mbStepNotifyPending = false;

	auto p = std::move(mpStepHandler);
	mpStepHandler = nullptr;

	if (p)
		p(!mbStepNotifyPendingBP);
}
